/*****************************************************************************
 COPYRIGHT

 (c) 1994-2004 by BebboSoft, Stefan "Bebbo" Franke, all rights reserved

 *****************************************************************************/

package de.bb.minissl;

/**
 * A small and fast implementation for BigInteger.
 * Like the String class, all  functions do not modify an existing BigInteger,
 * but return a new created BigInteger with the result.
 */
public class FastMath32
{
  /**
   * @param dst
   * @param src
   * @param len
   * /
  public static native void square_native(int[]dst, int[] src, int len);	
  /**
   * @param dst
   * @param a
   * @param b
   * @param len
   * /
  public static native void mul_native(int[] dst, int[] a, int[] b, int len);
  /**
   * @param t
   * @param mod
   * @param n0
   * @param ml
   * @return
   * /
  public static native boolean mont_native(int t[], int mod[], int n0, int ml);
  */
//  static long start;

  /**
   * calculate z**e mod n
   * @param z some value
   * @param mod the modulo
   * @param exp the exponent
   * @return
   */
  public static byte[] oddModPow(byte[] z, byte[] exp, byte[] mod)
  {
//    start = System.currentTimeMillis();
    // allocate memory for
    // 16 precomputed M3.squares
    // z, exp, mod
    // t1, t2
    int[][] data = new int[21][];
    int modLen = (mod.length + 3) / 4;
    int wordLen = modLen * 2;
    for (int i = 0; i < 20; ++i) {
      data[i] = new int[wordLen + 1];
    }
    copyByte2Int(data[16], z);
    copyByte2Int(data[17], exp);
    copyByte2Int(data[18], mod);
    while (data[18][modLen - 1] == 0) {
      --modLen;
    }

//    ////System.out.println("1:" + (System.currentTimeMillis() - start));

    int[] ir = oddModPow(data, modLen);

//    ////System.out.println("9:" + (System.currentTimeMillis() - start));
    /*    
     FastMathTest.dump(ir);

     IntImpl zi = new IntImpl(z);
     IntImpl expi = new IntImpl(exp);
     IntImpl modi = new IntImpl(mod);
     IntImpl res = zi.modPow(expi, modi);
     ////System.out.println(res.toHexString());
     */
    int i = 0;
    while (mod[i] == 0)
      ++i;
    byte r[] = new byte[mod.length - i];
    copyLong2Byte(r, ir);
    return r;
  }

  /**
   * Calculate this^exp % mod for odd modulus.
   * We assume all arrays are large enough to hold mod.len * 2 + 1
   * @param data
   * @param modLen
   * @return
   */
  private static int[] oddModPow(int data[][], int modLen)
  {
    int maxLen = modLen * 2;

    int z[] = data[16];
    int exp[] = data[17];
    int mod[] = data[18];

    // get the count of exponent bits
    int bits2Process = bitLength(exp, modLen);

//    ////System.out.println("2:" + (System.currentTimeMillis() - start));

    int n0 = modInverse32(mod[0]);
    ////System.out.println("modInverse of " + Integer.toHexString(mod[0]) + " is " + Integer.toHexString(n0));

    int modOffset = modLen << 5;
    int t1[] = data[0];

    // data[1] = z inverse 
    // IntImpl dis = shiftLeft(modOffset).mod(mod);
    ////System.out.println("z:" + new IntImpl(z, maxLen, true).toHexString());
    shiftLeft(data[1], z, modLen, modOffset);
    ////System.out.println("shl:" + new IntImpl(data[1], maxLen, true).toHexString());
    mod(data[1], mod, data[2], data[3], maxLen, modLen);
    ////System.out.println("modulo:" + new IntImpl(mod, maxLen, true).toHexString());
    ////System.out.println("shl mod modulo:" + new IntImpl(data[1], maxLen, true).toHexString());

    // m = 2**bits, where m*m = sqrt(ipr/2)    
    int bitsAtOnce = 1;
    for (int i = bits2Process / 2; i != 0; i = (i & 0xffff) >> 1)
      ++bitsAtOnce;
    bitsAtOnce /= 2; // sqrt
    bitsAtOnce += -1;
    if (bitsAtOnce == 0)
      bitsAtOnce = 1;
    // limit
    if (bitsAtOnce > 4)
      bitsAtOnce = 4;

    // size for array of precalculated powers of a
    int sqaSize = (1 << bitsAtOnce);

    --bits2Process;
    int currentBits = bits2Process & 31;
    int words2Process = bits2Process >>> 5;

    if (currentBits == 0) {
      currentBits = 32;
      --words2Process;
    }

    ////System.out.println("n:" + new IntImpl(data[1], maxLen, true).toHexString());

    boolean x[] = new boolean[sqaSize];
    x[0] = true;
    x[1] = true;

//    ////System.out.println("4:" + (System.currentTimeMillis() - start));

    t1 = data[1];

    if (bits2Process > 0) {

      int t2[] = data[0];
      ////System.out.println(exp[words2Process] + " - " + Integer.toHexString(32 - currentBits)); 
      int mask = (exp[words2Process] << (32 - currentBits));
      int ubits = 0, ui = 0, pass = 0;
      for (;;) {
        ////System.out.println(Integer.toHexString(mask) + " - " + Integer.toHexString(ui)); 
        if (++pass < 0) {
          int off = 1 << (pass - 1);
          square(data[off], t1, modLen);
          montgomery(data[off], mod, n0, modLen);
          x[off] = true;
          t2 = data[0];
          t1 = data[off];
        } else {
          square(t2, t1, modLen);
          montgomery(t2, mod, n0, modLen);
          t1 = t2;
          t2 = t1 == data[0] ? data[19] : data[0];
        }

        ui <<= 1;
        if (mask < 0)
          ui |= 1;
        mask <<= 1;

        if (ui != 0)
          ++ubits;
        if (ubits == bitsAtOnce) {
          if (ui > 0) {
            if (!x[ui]) {
              fill(bitsAtOnce, ui, x, data, mod, modLen, n0);
            }

            mul(t2, t1, data[ui], modLen);
            montgomery(t2, mod, n0, modLen);
            t1 = t2;
            t2 = t1 == data[0] ? data[19] : data[0];
          }
          ubits = 0;
          ui = 0;
        }

        //////System.out.println("r: " + new IntImpl(t1, maxLen, true).toHexString());

        if (--currentBits == 0) {
          if (words2Process-- == 0)
            break;
          currentBits = 16;
          mask = (exp[words2Process] & 0xffff) << 16;
        }
      }

      if (ubits != 0 && ui > 0) {
        if (!x[ui]) {
          fill(bitsAtOnce, ui, x, data, mod, modLen, n0);
        }
        mul(t2, t1, data[ui], modLen);
        montgomery(t2, mod, n0, modLen);
        t1 = t2;
      }
    }
    montgomery(t1, mod, n0, modLen);
    return t1;
  }

  private static void fill(int bitsAtOnce, int ui, boolean x[], int data[][],
      int mod[], int modLen, int n0)
  {
    ////System.out.println("-> fill");
    for (int s = bitsAtOnce - 2; s >= 0; --s) {
      int xui = ui >> s;
      if (!x[xui & 0xfe]) {
        square(data[xui & 0xfe], data[xui >> 1], modLen);
        montgomery(data[xui & 0xfe], mod, n0, modLen);
        x[xui & 0xfe] = true;
      }
      if (!x[xui]) {
        mul(data[xui], data[1], data[xui - 1], modLen);
        montgomery(data[xui], mod, n0, modLen);
        x[xui] = true;
      }
    }
    ////System.out.println("fill ->");
  }

  /**
   * @param src
   * @param dst
   */
  static void copyLong2Byte(byte[] dst, int[] src)
  {
    for (int i = 0, j = dst.length - 1; i < src.length; ++i) {
      int v = src[i];
      if (j < 0)
        break;
      dst[j--] = (byte) v;
      v >>>= 8;
      if (j < 0)
        break;
      dst[j--] = (byte) v;
      v >>>= 8;
      if (j < 0)
        break;
      dst[j--] = (byte) v;
      v >>>= 8;
      if (j < 0)
        break;
      dst[j--] = (byte) v;
    }
  }

  /**
   * @param src
   * @param dst
   */
  public static void copyByte2Int(int[] dst, byte[] src)
  {
    for (int i = 0, j = src.length - 1; i < dst.length && j >= 0; ++i) {
      int v = (src[j--] & 0xff);
      if (j >= 0)
        v |= ((src[j--] & 0xff) << 8);
      if (j >= 0)
        v |= ((src[j--] & 0xff) << 16);
      if (j >= 0)
        v |= ((src[j--] & 0xff) << 24);
      dst[i] = v;
    }
  }

  /**
   * sub WITHOUT handling underflows!!!!
   * @param res
   * @param a
   * @param b
   * @param bl
   */
  public static void sub(int[] res, int[] a, int[] b, int bl)
  {
    long carry = 0;
    int i;
    for (i = 0; i < bl; ++i) {
      carry += (a[i] & 0xffffffffL) - (b[i] & 0xffffffffL);
      res[i] = (int) carry;
      carry >>= 32;
    }
    for (; carry != 0 && i < bl; ++i) {
      carry += a[i] & 0xffffffffL;
      res[i] = (int) carry;
      carry >>= 32;
    }
  }

  /**
   * @param aIn
   * @param bIn
   * @param c
   * @param d
   * @param al
   * @param bl
   */
  public static void mod(int[] aIn, int[] bIn, int c[], int d[], int al, int bl)
  {
    int a[] = aIn;
    int b[] = bIn;

    while (bl > 0) {
      if (b[bl - 1] != 0)
        break;
      --bl;
    }

    int shift = 0;
    {
      int x = b[bl - 1];
      if (x < 0)
        shift = 31;
      else {
        while (x + x > 0) {
          x += x;
          ++shift;
        }
      }

      if (shift > 0) {
        shiftLeft(c, bIn, bl, shift);
        b = c;
        shiftLeft(d, aIn, al, shift);
        a = d;

        if (b[bl] != 0)
          ++bl;
        if (a[al] != 0)
          ++al;
      }
    }
    ////System.out.print("a: ");
    dump(a);
    ////System.out.print("b: ");
    dump(b);
    
    int divlen = bl - 1;

    // modMSW aligned that highest bit is zero and 2nd highest bit is set
    long modMSW = b[divlen] & 0xffffffffL;
    long modMSW32 = modMSW << 32;
    long modLSW = divlen > 0 ? b[divlen - 1] & 0xffffffffL : 0;
    for (int i = al - bl; i >= 0;) {
      long a0 = al == a.length ? 0 : a[al] & 0xffffffffL;
      long a01 = (a0 << 32) | (a[--al] & 0xffffffffL);

      // calculate:  c0 = a0|a1 / b0
      //      long c0 = a0 == modMSW ? 0xffffffffL : (a01 / modMSW) & 0xffffffffL;
      long c0;
      if (a0 == modMSW) {
        c0 = 0xffffffffL; // wirklich noetig?
      } else
      //if (a01 >= 0) 
      {
        c0 = (a01 / modMSW) & 0xffffffffL;
        //      } else {
        //        c0 = (((a01>>>1) / modMSW) << 1) + 1 ;
      }
      ////System.out.println("c0 = " + Long.toHexString(c0));

      if (c0 > 0) {
        if (al > 0 && modLSW != 0) {
          // a0|a1 -= c0*b0 // a0 part always gets zero!
          a01 -= c0 * modMSW;

          long c0b1 = c0 * modLSW;
          // add a2 => a0|a1|a2 - c0*b0|0          
          a01 = (a01 << 32) | (a[al - 1] & 0xffffffffL);
          while ((a01 >= 0 && (c0b1 < 0 || (c0b1 >= 0 && a01 < c0b1)))
              || (a01 < 0 && c0b1 < 0 && a01 < c0b1)) {
            --c0;
            long t = a01 + modMSW32;
            c0b1 -= modLSW;
            if (c0b1 < 0 && a01 < 0 && t >= 0)
              break;
            a01 = t;
          }
        }
        ////System.out.println("c0'= " + Long.toHexString(c0));

        if (c0 > 0) {
          // start of mulSub
          ////System.out.print("mulsub0: ");
          dump(a);
          
          int ai = al - divlen;
          long carry = 0;
          for (int bi = 0; bi <= divlen; ++bi, ++ai) {
            long p = (b[bi] & 0xffffffffL) * c0;
            ////System.out.print(Long.toHexString(p) + " = " + Long.toHexString(b[bi] & 0xffffffffL) + " * " + Long.toHexString(c0) + " -> ");
            carry += (b[bi] & 0xffffffffL) * c0;
            ////System.out.println(Long.toHexString(carry));
            int t = a[ai];
            a[ai] = t - (int) carry;
            carry = (carry >>> 32)
                + (((int) carry + 0x80000000) > (t + 0x80000000) ? 1 : 0);
          } 
          ////System.out.println(Long.toHexString(carry));
          if (carry != 0) {
            carry = a[ai] - carry;
            a[ai] = (int) carry;
          }
          if (carry < 0)
          //if (mulSub(al, divlen, c0, b, a)) // got negative result?
          {
            ////System.out.print("mulsub1: ");
            dump(a);
            ////System.out.println("c0:" + c0);
            --c0;

            carry = 0;
            ai = al - divlen;
            for (int bi = 0; bi <= divlen; bi++) {
              carry += (b[bi] & 0xffffffffL);
              carry += (a[ai] & 0xffffffffL);
              a[ai++] = (int) carry;
              carry >>>= 32;
            }
            if (carry > 0)
              a[ai] = a[ai] + (int) carry;
          }
          ////System.out.print("mulsub2: ");
          dump(a);          
        }
      }
      --i; // c[i--] += (int) c0;
    }

    if (shift > 0) {
      shiftRight(aIn, a, aIn.length, shift);
    }
  }

  /**
   * @param dst
   * @param src
   * @param len
   */
  public static void square(int[] dst, int[] src, int len)
  {
    ////System.out.println("square: " + new IntImpl(src, len, true).toHexString());
    // calc the M3.squares a1*a1, a2*a2, ... - also erases the buffer
    int i = len - 1;
    int j = i + len;
    dst[j] = 0;
    for (; i >= 0; --i) {
      long p = src[i] & 0xffffffffL;
      p *= p;
      dst[j] = (int) (p >>> 32);
      --j;
      dst[j] = (int) p;
      --j;
    }

    // add the mixed terms
    for (i = 0; i < len; ++i) {
      int l = src[i];
      long carry = 0;
      j = i + 1;
      int k = i + j;
      for (; j < len; ++j, ++k) {
        long m = (l & 0xffffffffL) * (src[j] & 0xffffffffL);
        carry += m + m + (dst[k] & 0xffffffffL);
        dst[k] = (int) carry;
        //if (carry + 0x8000000000000000L < m + 0x8000000000000000L)
        if ((int)(m >>> 32) < 0) // overflow is only caused by m!
          carry = 0x100000000L | (carry >>> 32);
        else
          carry = (carry >>> 32);
      }
//      for (; carry != 0; ++k) {
      for (; (int)carry != 0; ++k) {
        carry += (dst[k] & 0xffffffffL);
        dst[k] = (int) carry;
        carry >>>= 32;
      }
    }
    ////System.out.println("square= " + new IntImpl(dst, len, true).toHexString());
  }

  /**
   * @param dst
   * @param a
   * @param b
   * @param len
   */
  public static void mul(int[] dst, int[] a, int[] b, int len)
  {
    ////System.out.println("mul:a " + new IntImpl(a, len, true).toHexString());    
    ////System.out.println("mul:b " + new IntImpl(b, len, true).toHexString());    
    long carry = 0;
    int temp = a[0];
    int i = 0;
    for (; i < len; ++i) {
      carry += (b[i] & 0xffffffffL) * (temp & 0xffffffffL);
      dst[i] = (int) carry;
      carry >>>= 32;
    }
    dst[i] = (int) carry;

    for (int q = 1; q < len; ++q) {
      carry = 0;
      temp = a[q];
      i = q;
      for (int bi = 0; bi < len; ++bi) {
        carry += (b[bi] & 0xffffffffL) * (temp & 0xffffffffL);
        carry += dst[i] & 0xffffffffL;

        dst[i++] = (int) carry;
        carry >>>= 32;
      }
      dst[i] = (int) carry;
    }
    ////System.out.println("mul=  " + new IntImpl(dst, len, true).toHexString());    
  }
  public static void mul2(int[] dst, int[] a, int[] b, int len)
  {
    ////System.out.println("mul:a " + new IntImpl(a, len, true).toHexString());    
    ////System.out.println("mul:b " + new IntImpl(b, len, true).toHexString());    
    long carry = 0;
    long temp = a[0] & 0xffffffffL;
    int i = 0;
    for (; i < len; ++i) {
      carry += (b[i] & 0xffffffffL) * temp;
      dst[i] = (int) carry;
      carry >>>= 32;
    }
    dst[i] = (int) carry;

    for (int q = 1; q < len; ++q) {
      carry = 0;
      temp = a[q] & 0xffffffffL;
      i = q;
      for (int bi = 0; bi < len; ++bi) {
        carry += (b[bi] & 0xffffffffL) * temp;
        carry += dst[i] & 0xffffffffL;

        dst[i++] = (int) carry;
        carry >>>= 32;
      }
      dst[i] = (int) carry;
    }
    ////System.out.println("mul=  " + new IntImpl(dst, len, true).toHexString());    
  }

  /**
   * @param dst
   * @param src
   * @param len
   * @param shift
   */
  static void shiftLeft(int[] dst, int[] src, int len, int shift)
  {
    long carry = 0;
    int j = shift >>> 5;
    shift &= 31;
    for (int i = 0; i < len; ++i) {
      carry >>>= 32;
      carry |= (src[i] & 0xffffffffL) << shift;
      dst[j++] = (int) carry;
    }
    if ((carry >>= 32) != 0)
      dst[j] = (int) carry;
  }

  static void shiftRight(int[] dst, int[] src, int len, int shift)
  {
    long carry = 0;
    int j = len - (shift >>> 5) - 1;
    shift &= 31;
    for (int i = len - 1; j >= 0; --i) {
      carry <<= 32;
      carry |= (src[i] & 0xffffffffL);
      dst[j--] = (int) (carry >>> shift);
    }
  }

  /**
   * Return the count of used bits.
   * @param exp
   * @param len
   * @return
   */
  static int bitLength(int[] exp, int len)
  {
    while (--len >= 0) {
      if (exp[len] != 0)
        break;
    }
    if (len < 0)
      return 0;
    int w = exp[len];
    int n = 0;
    for (; w != 0; w >>>= 1)
      ++n;
    return len * 32 + n;
  }

  /**
   * Internal function to calculate a modInverse for the lowest int.
   * @param _x lowest int of n.
   * @return A result where: x*result mod 2^16 = 1
   */
  static int modInverse32(int _x)
  {
    //    int z = 0;
    long x = _x & 0xffffffffL;
    long y = 1;
    long s = 2;
    for (int i = 2; i <= 32; ++i) {
      long ss = s << 1;
      if (s < (x * y & (ss - 1)))
        y += s;
      s = ss;
      //      if ((x * y & s - 1) != 1)  y /= z;
    }
    return -(int) y;
  }

  /**
   * @param t
   * @param mod
   * @param ml
   * @param n0
   */
  static void montgomery(int[] t, int[] mod, int n0, int ml)
  {
    ////System.out.println("mont: " + new IntImpl(t, ml, true).toHexString());    
    /** /
     short[] mods = new short[mod.length * 2];
     short[] ts = new short[t.length * 2];
     copyInt2Short(mods, mod);
     copyInt2Short(ts, t);
     montgomery(ts, mods, ml * 2, (short) n0);
     copyShort2Int(t, ts);

     /**/
     /** /
    // t = (t + (t*n' mod r) * n )
    long carry = 0;
    for (int i = 0; i < ml; ++i) {
      long m = (n0 * t[i]) & 0xffffffffL;
      int k = i;
      for (int j = 0; j < ml; ++j, ++k) {
        carry += (t[k] & 0xffffffffL) + m * (mod[j] & 0xffffffffL);
        t[k] = (int) carry;
        carry >>>= 32;
      }
      for (; carry != 0 && k < ml2; ++k) {
        carry += (t[k] & 0xffffffffL);
        t[k] = (int) carry;
        carry >>>= 32;
      }
    }
  
    boolean sub = carry != 0;  
    /**/

    boolean sub = mont(t, mod, n0, ml);
    
    int ml2 = ml + ml;
    // t = t / r              = shift right
    {
      int i = 0;
      for (; i < ml; ++i)
        t[i] = t[i + ml];
      for (; i < ml2; ++i)
        t[i] = 0;
    }

    // boolean sub = carry != 0;
    if (!sub) {
      int i = ml - 1;
      for (; i >= 0; --i) {
        if (t[i] != mod[i])
          break;
      }
      sub = i < 0 || t[i] + 0x80000000 > mod[i] + 0x80000000;
//      sub = i < 0 || t[i] - mod[i] < 0;
    }
    if (sub) {
      sub(t, t, mod, ml);
    }
    /**/
    ////System.out.println("mont= " + new IntImpl(t, ml, true).toHexString());    
  }
  
  /**
   * @param t
   * @param mod
   * @param n0
   * @param ml
   * @return
   */
  public static boolean mont(int[] t, int[] mod, int n0, int ml)
  {
    // t = (t + (t*n' mod r) * n )
    int ml2 = ml + ml;
    long carry = 0;
    for (int i = 0; i < ml; ++i) {
      long m = (n0 * t[i]) & 0xffffffffL;
      int k = i;
      for (int j = 0; j < ml; ++j, ++k) {
        carry += (t[k] & 0xffffffffL) + (m) * (mod[j] & 0xffffffffL);
        t[k] = (int) carry;
        carry >>>= 32;
      }
      for (; carry != 0 && k < ml2; ++k) {
        carry += (t[k] & 0xffffffffL);
        t[k] = (int) carry;
        carry >>>= 32;
      }
    }
    return carry != 0;
  }  
  /**/
  static void dump(int[] src)
  {
    if (true) return;
    StringBuffer sb = new StringBuffer();
    for (int i = src.length - 1; i >= 0; --i) {
      int v = src[i];
      sb.append(Integer.toHexString((v>>28)&0xf));
      sb.append(Integer.toHexString((v>>24)&0xf));
      sb.append(Integer.toHexString((v>>20)&0xf));
      sb.append(Integer.toHexString((v>>16)&0xf));
      sb.append(Integer.toHexString((v>>12)&0xf));
      sb.append(Integer.toHexString((v>>8)&0xf));
      sb.append(Integer.toHexString((v>>4)&0xf));
      sb.append(Integer.toHexString((v)&0xf));
    }
    String s = sb.toString();
    while(s.length() > 1 && s.charAt(0) == '0')
      s = s.substring(1);
    System.out.println(s);
  }

  public static int[] hex2Int(String string)
  {    
    int len = (string.length() + 7) /8;
    int r[] = new int[len];
    for (int k = 0, p = string.length(); p > 0; ++k, p-=8) {
      int c = string.charAt(p - 1);
      if (c >= 'a') c -= 0x20;
      int v = c > '9' ? c - 'A' + 10 : c - '0';
      int n = v;
      if (p > 1) {
        c = string.charAt(p - 2);
        if (c >= 'a') c -= 0x20;
        v = c > '9' ? c - 'A' + 10 : c - '0';
        n |= v<<4;
        if (p > 2) {
          c = string.charAt(p - 3);
          if (c >= 'a') c -= 0x20;
          v = c > '9' ? c - 'A' + 10 : c - '0';
          n |= v<<8;
          if (p > 3) {
            c = string.charAt(p - 4);
            if (c >= 'a') c -= 0x20;
            v = c > '9' ? c - 'A' + 10 : c - '0';
            n |= v<<12;
            if (p > 4) {
              c = string.charAt(p - 5);
              if (c >= 'a') c -= 0x20;
              v = c > '9' ? c - 'A' + 10 : c - '0';
              n |= v<<16;
              if (p > 5) {
                c = string.charAt(p - 6);
                if (c >= 'a') c -= 0x20;
                v = c > '9' ? c - 'A' + 10 : c - '0';
                n |= v<<20;
                if (p > 6) {
                  c = string.charAt(p - 7);
                  if (c >= 'a') c -= 0x20;
                  v = c > '9' ? c - 'A' + 10 : c - '0';
                  n |= v<<24;
                  if (p > 7) {
                    c = string.charAt(p - 8);
                    if (c >= 'a') c -= 0x20;
                    v = c > '9' ? c - 'A' + 10 : c - '0';
                    n |= v<<28;
                  }
                }
              }
            }
          }
        }
      }
      r[k] = n;
    }
    
    return r;
  }
  
}